/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/impl/lib/absolute_fqdn.hh"
#include "src/impl/lib/exception.hh"

#include <utility>

namespace Fred {
namespace Zone {
namespace Impl {
namespace Lib {

namespace {

Fqdn fqdn_checked_for_length(Fqdn fqdn);

class FqdnView
{
public:
    explicit constexpr FqdnView(const char* begin, const char* end) noexcept
        : begin_{begin},
          end_{end}
    { }
    template <std::size_t n>
    explicit constexpr FqdnView(const char (&str)[n]) noexcept
        : begin_{str},
          end_{str + n - 1}
    {
        static_assert(0 < n, "string must occupy at least one character");
    }
    explicit FqdnView(const Fqdn& fqdn) noexcept
        : FqdnView{*fqdn}
    { }
    constexpr FqdnView(const FqdnView&) noexcept = default;
    constexpr FqdnView(FqdnView&&) noexcept = default;
    constexpr FqdnView& operator=(const FqdnView&) noexcept = default;
    constexpr FqdnView& operator=(FqdnView&&) noexcept = default;
    constexpr bool is_within(const FqdnView& domain) const noexcept
    {
        return is_first_within_second(make_relative(*this), make_relative(domain));
    }
private:
    explicit FqdnView(const std::string& str)
        : begin_{&*std::begin(str)},
          end_{&*std::end(str)}
    { }
    constexpr auto length() const noexcept
    {
        return end_ - begin_;
    }
    constexpr bool is_empty() const noexcept
    {
        return this->length() <= 0;
    }
    constexpr char front() const noexcept
    {
        return *begin_;
    }
    constexpr char back() const noexcept
    {
        return end_[-1];
    }
    constexpr bool is_same(const char* rhs) const noexcept
    {
        for (const char* lhs = begin_; lhs < end_; ++lhs, ++rhs)
        {
            if (*lhs != *rhs)
            {
                return false;
            }
        }
        return true;
    }
    template <typename T>
    constexpr FqdnView tail(T length) const noexcept
    {
        return FqdnView{end_ - length, end_};
    }
    const char* begin_;
    const char* end_;
    friend constexpr bool is_first_within_second(const FqdnView& first, const FqdnView& second) noexcept
    {
        if (second.is_empty())
        {
            return true;
        }
        if (first.is_empty())
        {
            return false;
        }
        if (first.length() == second.length())
        {
            return first == second;
        }
        const auto min_length_of_subdomain = second.length() + 2;
        if (first.length() < min_length_of_subdomain)
        {
            return false;
        }
        if (first.tail(second.length() + 1).front() != '.')
        {
            return false;
        }
        return first.tail(second.length()) == second;
    }
    friend constexpr FqdnView make_relative(const FqdnView& fqdn) noexcept
    {
        return fqdn.is_empty() || (fqdn.back() != '.') ? fqdn
                                                       : FqdnView{fqdn.begin_, fqdn.end_ - 1};
    }
    friend constexpr bool is_absolute(const FqdnView& fqdn)
    {
        return !fqdn.is_empty() && (fqdn.back() == '.');
    }
    friend Fqdn fqdn_checked_for_length(Fqdn fqdn)
    {
        constexpr auto max_fqdn_length = 255;
        if (max_fqdn_length < (make_relative(FqdnView{fqdn}).length() + 1))
        {
            throw_runtime_error([](){ return "fqdn must be at most 255 characters long"; });
        }
        return fqdn;
    }
    friend constexpr bool operator==(const FqdnView& lhs, const FqdnView& rhs) noexcept
    {
        return (lhs.length() == rhs.length()) && lhs.is_same(rhs.begin_);
    }
    friend constexpr bool operator<(const FqdnView& lhs, const FqdnView& rhs) noexcept
    {
        const char* lhs_top = lhs.begin_;
        const char* rhs_top = rhs.begin_;
        while ((lhs_top != lhs.end_) && (rhs_top != rhs.end_))
        {
            if (*lhs_top < *rhs_top)
            {
                return true;
            }
            if (*rhs_top < *lhs_top)
            {
                return false;
            }
            ++lhs_top;
            ++rhs_top;
        }
        return (lhs_top == lhs.end_) && (rhs_top != rhs.end_);
    }
};

static_assert(make_relative(FqdnView{"."}) == FqdnView{""}, "terminating '.' must be removed");
static_assert(make_relative(FqdnView{"cz."}) == FqdnView{"cz"}, "terminating '.' must be removed");
static_assert(make_relative(FqdnView{"domain.cz"}) == FqdnView{"domain.cz"}, "must be equal");

static_assert(FqdnView{"sub.domain.cz"}.is_within(FqdnView{"domain.cz"}), "must be within");
static_assert(FqdnView{"sub.domain.cz."}.is_within(FqdnView{"domain.cz"}), "must be within");
static_assert(FqdnView{"sub.domain.cz"}.is_within(FqdnView{"domain.cz."}), "must be within");
static_assert(FqdnView{"sub.domain.cz."}.is_within(FqdnView{"domain.cz."}), "must be within");

static_assert(FqdnView{"sub.domain.cz."}.is_within(FqdnView{""}), "must be within");
static_assert(FqdnView{"sub.domain.cz."}.is_within(FqdnView{"."}), "must be within");

static_assert(FqdnView{""}.is_within(FqdnView{""}), "must be within");
static_assert(FqdnView{"."}.is_within(FqdnView{""}), "must be within");
static_assert(FqdnView{""}.is_within(FqdnView{"."}), "must be within");
static_assert(FqdnView{"."}.is_within(FqdnView{"."}), "must be within");

static_assert(FqdnView{"domain.cz"}.is_within(FqdnView{"domain.cz"}), "must be within");

static_assert(!FqdnView{".domain.cz"}.is_within(FqdnView{"domain.cz"}), "cannot be within");
static_assert(!FqdnView{"domain.cz"}.is_within(FqdnView{"sub.domain.cz"}), "cannot be within");
static_assert(!FqdnView{""}.is_within(FqdnView{"domain.cz"}), "cannot be within");

static_assert(is_absolute(FqdnView{"."}), "must be absolute");
static_assert(is_absolute(FqdnView{"domain.cz."}), "must be absolute");
static_assert(!is_absolute(FqdnView{""}), "cannot be absolute");
static_assert(!is_absolute(FqdnView{"domain.cz"}), "cannot be absolute");

std::string make_absolute_fqdn(const Fqdn& fqdn)
{
    if (is_absolute(FqdnView{fqdn}))
    {
        return *fqdn;
    }
    return (*fqdn) + ".";
}

}//namespace Fred::Zone::Impl::Lib::{anonymous}

AbsoluteFqdn::AbsoluteFqdn(Fqdn fqdn)
    : fqdn_{fqdn_checked_for_length(std::move(fqdn))}
{ }

Fqdn AbsoluteFqdn::get_fqdn() const
{
    return Fqdn{make_absolute_fqdn(fqdn_)};
}

bool AbsoluteFqdn::is_within(const AbsoluteFqdn& domain) const noexcept
{
    return FqdnView{fqdn_}.is_within(FqdnView{domain.fqdn_});
}

bool operator<(const AbsoluteFqdn& lhs, const AbsoluteFqdn& rhs) noexcept
{
    return make_relative(FqdnView{lhs.fqdn_}) < make_relative(FqdnView{rhs.fqdn_});
}

AbsoluteFqdn make_absolute(Fqdn fqdn)
{
    return AbsoluteFqdn{std::move(fqdn)};
}

}//namespace Fred::Zone::Impl::Lib
}//namespace Fred::Zone::Impl
}//namespace Fred::Zone
}//namespace Fred
